/***************************************************************************************************
*             C# sample for the usage of Episodic TopoART (class Fast_Episodic_TopoART)            *
****************************************************************************************************
*                           Created by Marko Tscherepanow, 20 October 2012                         *
***************************************************************************************************/

// Compile and run from the console: dotnet run -p Episodic_TopoART_sample1.csproj

using System;
using System.IO;
using System.Reflection;
using System.Text.RegularExpressions;
using LibTopoART;
using System.Globalization;

namespace LibTopoART_samples
{
	/// <summary>
	/// Episodic clustering sample using synthetic two-dimensional data. [C#]
	/// <para>
	/// Like in Section 4.1 of "Marko Tscherepanow, Sina Kühnel, and Sören Riechers (2012). Episodic 
	/// Clustering of Data Streams Using a Topology-Learning Neural Network. In Proceedings of the European 
	/// Conference on Artificial Intelligence (ECAI), Workshop on Active and Incremental Learning (AIL), 
	/// pp. 24-29. Montpellier, France.", an Episodic TopoART network is trained with the well-known Two 
	/// Spirals dataset. Due to the incorporation of temporal information during learning, Episodic TopoART 
	/// is capable of creating two clusters each representing one spiral in an unsupervised way. 
	/// These clusters are formed by the nodes of module b (ETA b).
	/// </para>
	/// <para>The resulting network can be visualised using the R script <c>ShowEpisodicTopoARTResults.R</c> 
	/// provided in the subfolder <c>R</c>.
	/// </para>
	/// </summary>
	class Episodic_TopoART_sample1
	{
	    static void Main() {
			
			// Dataset
			const string Dataset		=	"../../../../../data/TwoSpirals_dataset.txt";	
			// Destination directory for trained networks
			const string NetworkPath	=	"../../../../../results/networks/";

			const long SampleNumber		=	194;
			const long InputDimension	=	2;

			decimal[][] InputArray		=	new decimal[SampleNumber][];
			StreamReader DatasetFile;
			long i;
			string line;
			Fast_Episodic_TopoART feta;

			// Set working directory to assembly directory
            Directory.SetCurrentDirectory(Path.GetDirectoryName(new Uri(Assembly.GetEntryAssembly().Location).LocalPath));

			// Load dataset
			using(DatasetFile = new StreamReader(File.Open(Dataset, FileMode.Open))) {
				for(i = 0; i < SampleNumber; ++i) {
					line = DatasetFile.ReadLine();
					string[] numbers = Regex.Split(line, @"\s+");
					InputArray[i] = new decimal[InputDimension]; 
					InputArray[i][0] = Decimal.Parse(numbers[1], NumberStyles.Float, System.Globalization.CultureInfo.InvariantCulture);
					InputArray[i][1] = Decimal.Parse(numbers[2], NumberStyles.Float, System.Globalization.CultureInfo.InvariantCulture);
					InputArray[i][0] = (InputArray[i][0] + 7.0m) / 14m;
					InputArray[i][1] = (InputArray[i][1] + 7.0m) / 14m;
				}
			}
			
			// Initialise a Episodic TopoART network with appropriate parameter values (taken from the paper
			// given above)
			feta = new Fast_Episodic_TopoART(InputDimension, 2, 0.69m, 200);
			feta.Beta_sbm	= 0.0m;
			feta.Phi		= 1;
			feta.Tau		= 200;
					
			// Train network
			for(i = 0; i < SampleNumber; ++i) 
				feta.Learn(InputArray[i]);
			
			// Determine clusters
			feta.ComputeClusterIDs();
			
			// Save network in human readable form
			feta.SaveText(NetworkPath + "Fast_Episodic_TopoART_TwoSpirals_dataset.txt");

			// Save network in binary form
			feta.Save(NetworkPath + "Fast_Episodic_TopoART_TwoSpirals_dataset.feta");
		}
	}
}