(***************************************************************************************************
*              Extended F# sample for the usage of TopoART-AM (class Fast_TopoART_AM)              *
****************************************************************************************************
*                            Created by Marko Tscherepanow, 14 March 2020                          *
***************************************************************************************************)

// Compile and run from the console: dotnet run -p TopoART-AM_sample2.fsproj

// macOS: If libgdiplus cannot be loaded, an installation of mono-libgdiplus via homebrew may be required.
//        Alternatively, the nuget package runtime.osx.10.10-x64.CoreCompat.System.Drawing can be added to the 
//        project, for instance, by entering the following command in the project directory: 
//        dotnet add package runtime.osx.10.10-x64.CoreCompat.System.Drawing 
// Raspbian, Ubuntu: If libgdiplus cannot be loaded, its installation using the package manager may be required.

/// <summary>
/// Learning of bidirectional associations between images. [F#]
/// <para>
/// Similar to Section 4.2 of "Marko Tscherepanow, Marco Kortkamp, and Marc Kammer (2011). A Hierarchical ART
/// Network for the Stable Incremental Learning of Topological Structures and Associations from Noisy Data. 
/// Neural Networks 24(8): 906-916. Elsevier.", a TopoART-AM network is trained with real-world image data.
/// There are two kinds of images grouped into owners and objects. TopoART-AM learns a bi-directional mapping
/// between images of these two groups. Each image has a size of about 34500 pixels. As each pixel comprises 3 
/// color channels (RGB) and the input vector encompasses a key from each group, the total length of the input 
/// vector is larger than 200,000. After finishing training, recall is performed for a single input stimulus of 
/// each group. The recall results are saved in the folder <c>results/recall/ObjectsOwners_dataset_recall_results</c>.
/// </para>
///</summary>
module LibTopoART_samples.TopoART_AM_sample2

open LibTopoART
open System
open System.Drawing
open System.Drawing.Imaging
open System.Globalization
open System.IO
open System.Reflection
open System.Runtime.InteropServices

// Dataset (containing training and test images)
let DatasetPath = "../../../../../data/ObjectsOwners_dataset/";

// Destination directory for the recall results
let ResultPath = "../../../../../results/recall/ObjectsOwners_dataset_recall_results/"

// Destination directory for trained networks
let NetworkPath = "../../../../../results/networks/"

// Key sizes
let Key1Width = 214
let Key1Height = 161
let Key1Len = int64 <| Key1Width * Key1Height * 3; // Size of key 1 (object images)
let Key2Width = 172
let Key2Height = 201
let Key2Len = int64 <| Key2Width * Key2Height * 3; // Size of key 1 (owner images)

// TopoART-AM parameters
let rho_a = 0.80m
let beta_sbm = 0.8m
let phi = 5L
let tau = 136L   // one fifth of the training dataset size (entries in ObjectMap * TrainImageNum)

// Application parameters
let ObjectNum = 20
let OwnerNum = 6
let TrainImageNum = 20
let TestImageNum = 5
let MaxTrainIterations = 25

// m to n mapping from owners to objects (some objects are shared by multiple owners)
let ObjectMap = [| [| 1; 2; 3; 9; 17; 19 |];             // objects of owner 1
                   [| 3; 4; 9; 15; 17 |];                // objects of owner 2
                   [| 1; 3; 7; 10; 11; 12; 16; 17 |];    // objects of owner 3
                   [| 3; 5; 17; 18; 20 |];               // objects of owner 4
                   [| 3; 8; 14; 19; 20 |];               // objects of owner 5
                   [| 3; 5; 10; 13; 16 |] |]             // objects of owner 6

(*--------------------------------------------------------------------------------------------------
-                                           Helper functions                                       -
--------------------------------------------------------------------------------------------------*)

let LoadSingleImage (path : string) =
    let bitmap = new Bitmap(path)
    if bitmap.PixelFormat <> System.Drawing.Imaging.PixelFormat.Format24bppRgb then
        printf "Incompatible pixel format"
        null
    else 
        let width = bitmap.Width
        let height = bitmap.Height

        if (width = Key1Width && height = Key1Height) || (width = Key2Width && height = Key2Height) then
            let PixelNumber = width * height
            let ImageSize = PixelNumber * 3
            let ImageArray = Array.create ImageSize 0uy
            let data = bitmap.LockBits(new Rectangle(new Point(0, 0), new Size(width, height)), ImageLockMode.ReadOnly, PixelFormat.Format24bppRgb)

            let mutable ptr = data.Scan0
            let mutable offset = 0
            let w3 = 3 * width
            for h = 0 to height - 1 do
                Marshal.Copy(ptr, ImageArray, offset, w3)
                ptr <- ptr + nativeint data.Stride
                offset <- offset + w3
            bitmap.UnlockBits(data)

            ImageArray
        else
            printf "loading %s failed\n" path
            null

let LoadImages index num subpath =
    let CommonPath = DatasetPath + subpath + "image_"
    let images = Array.init<byte[]> num (fun _ -> null)
    for i = 1 to num do
        let IndividualPath = CommonPath + i.ToString("D2") + ".jpg"
        images.[i - 1] <- LoadSingleImage IndividualPath
    images

let LoadImageArray sets subpath =
    let ImageArray = Array.init<byte[][]> sets (fun _ -> null)
    for i = 1 to sets do
        let path = subpath + i.ToString("D2") + "/"
        ImageArray.[i - 1] <- LoadImages i TrainImageNum path
    ImageArray

let LoadObjectImage (object : int) (image : int) = 
    LoadSingleImage <| DatasetPath + "test/objects/object_" + object.ToString("D2") + "/" + "image_" + image.ToString("D2") + ".jpg"

let LoadOwnerImage (owner : int) (image : int) = 
    LoadSingleImage <| DatasetPath + "test/owners/owner_" + owner.ToString("D2") + "/" + "image_" + image.ToString("D2") + ".jpg"

let Train (tam : IFast_TopoART_AM) map (objects : byte[][][]) (owners : byte[][][]) =
    let mutable TrainingSteps = 0
    let mutable stop = false
    let mutable i = 0
    while not stop && i < MaxTrainIterations do
        tam.ResetAdaptationState()
        i <- i + 1
        for owner = 1 to ObjectMap.Length do
            for ObjectIndex = 1 to ObjectMap.[owner - 1].Length do
                let object = ObjectMap.[owner - 1].[ObjectIndex - 1]
                for ImageIndex = 0 to TrainImageNum - 1 do
                    tam.Learn(objects.[object - 1].[ImageIndex], owners.[owner - 1].[ImageIndex])
                    TrainingSteps <- TrainingSteps + 1
            printf "."
        stop <- (tam.GetAdaptationState(0.001m) &&& AdaptationState.ANY_PERMANENT_ADAPTATION_MASK) = AdaptationState.NO_ADAPTATION
    printf "\n"
    TrainingSteps

let RecallImageName path (iteration : int) =
    path + "recall_image_" + iteration.ToString("D2") + ".jpg"

let RecallLoop min_F3_activation (tam : IFast_Associative_recall) =
    let mutable stop = false
    let mutable iteration = 1
    let mutable results = []
    while not stop do 
        let (success, result : byte[], activation : decimal) = tam.RecallStep()
        stop <- not success || activation < min_F3_activation
        if not stop then results <- (iteration, result, activation) :: results
        iteration <- iteration + 1
    results

let SaveActivations (activations : decimal list) path =
    use file = new StreamWriter(path + "activations.txt")
    for activation in activations do
        file.WriteLine(activation.ToString(CultureInfo.InvariantCulture))

let SaveRecallImage (image : byte[]) width height iteration path =
    let bitmap = new Bitmap(width, height, PixelFormat.Format24bppRgb)
    let data = bitmap.LockBits(new Rectangle(new Point(0, 0), new Size(width, height)), ImageLockMode.WriteOnly, PixelFormat.Format24bppRgb)

    let mutable ptr = data.Scan0
    let mutable offset = 0
    let w3 = 3 * width
    for h = 0 to height - 1 do
        Marshal.Copy(image, offset, ptr, w3)
        ptr <- ptr + nativeint data.Stride
        offset <- offset + w3

    bitmap.UnlockBits(data)

    if iteration > 0 then
        let file = RecallImageName path iteration
        bitmap.Save(file, ImageFormat.Jpeg)
    else
        bitmap.Save(path + "stimulus.jpg", ImageFormat.Jpeg)

let SaveResults (results : (int * byte [] * decimal) list) width height path =
    let mutable activations = []
    for (iteration, result, activation) in results do
        SaveRecallImage result width height iteration path
        activations <- activation :: activations

    let mutable iteration = results.Length + 1
    let mutable file = RecallImageName path iteration
    while File.Exists(file) do
        File.Delete(file)
        iteration <- iteration + 1
        file <- RecallImageName path iteration

    SaveActivations activations path

let RecallObjects owner image min_F3_activation (tam : IFast_Associative_recall) =
    let stimulus = LoadOwnerImage owner image
    let path = ResultPath + "key1/"
    SaveRecallImage stimulus Key2Width Key2Height 0 path

    // Start recall time measuring
    let recall_start = DateTime.Now

    let F3_nodes = tam.BeginRecallKey1(stimulus)
    let results = RecallLoop min_F3_activation tam
    tam.EndRecall()

    // Stop recall time measuring
    let recall_end = DateTime.Now

    // Output the required time
    let recall_time = recall_end - recall_start
    printf "Time for recalling key 1: %s\n" <| recall_time.ToString()

    // Save results
    SaveResults results Key1Width Key1Height path

let RecallOwners object image min_F3_activation (tam : IFast_Associative_recall) =
    let stimulus = LoadObjectImage object image
    let path = ResultPath + "key2/"
    SaveRecallImage stimulus Key1Width Key1Height 0 path

    // Start recall time measuring
    let recall_start = DateTime.Now

    let F3_nodes = tam.BeginRecallKey2(stimulus)
    let results = RecallLoop min_F3_activation tam
    tam.EndRecall()

    // Stop recall time measuring
    let recall_end = DateTime.Now

    // Output the required time
    let recall_time = recall_end - recall_start
    printf "Time for recalling key 2: %s\n" <| recall_time.ToString()

    // Save results
    SaveResults results Key2Width Key2Height path

(*--------------------------------------------------------------------------------------------------
-                                             Main program                                         -
--------------------------------------------------------------------------------------------------*)

[<EntryPoint>]
let main args =

    // Set working directory to assembly directory
    let cb = new Uri(Assembly.GetEntryAssembly().Location)
    Directory.SetCurrentDirectory(Path.GetDirectoryName(cb.LocalPath))

    printf "Load training images\n"
    let TrainingObjectImages = LoadImageArray ObjectNum "train/objects/object_"
    let TrainingOwnerImages = LoadImageArray OwnerNum "train/owners/owner_"

    // Create TopoART-AM network (with only a single module in order to accelerate computations)
    let tam = new Fast_TopoART_AM(Key1Len, Key2Len, 1L, rho_a)

    let FilePrefix = "Fast_TopoART-AM"
    let FileSuffix = "ftam"

    tam.Beta_sbm <- beta_sbm
    tam.Tau <- tau
    tam.Phi <- phi

    // Start training time measuring
    let training_start = DateTime.Now

    // Train
    let training_steps = Train tam ObjectMap TrainingObjectImages TrainingOwnerImages

    // Stop training time measuring
    let training_end = DateTime.Now

    // Output the required time
    let training_time = training_end - training_start
    printf "Time for training: %s\n" <| training_time.ToString()

    // Bidirectional recall of objects and owners based on images from the test set

    // The number if images that are recalled is controlled by the parameter min_F3_activation;
    // the lower the activation of an F3 node the less accurate is the respectively recalled 
    // image. Due to the similarity of the images, only a small fraction of the theoretically 
    // possible interval of [0, 1] can be used here.

    // The indexes may be changed in order to use different test stimuli. There are five test 
    // images per object and per owner.

    // Recall key 1 (objects) by presenting a stimulus key 2 (owner)
    RecallObjects 3 3 0.975m tam   // key 2: owner 3; test image: 3; recall threshold: 0.975

    // Recall key 2 (owners) by presenting a stimulus key 1 (object)
    RecallOwners 20 4 0.970m tam   // key 1: object 20; test image: 4; recall threshold: 0.970

    printf "Save network\n"

    // Save network in human readable form
    // tam.SaveText(NetworkPath + FilePrefix + "_ObjectsOwners_dataset.txt");

    // Save network in binary form
    tam.Save(NetworkPath + FilePrefix + "_ObjectsOwners_dataset." + FileSuffix);

    0
