/***************************************************************************************************
*             C# sample for the usage of TopoART (classes TopoART and Fast_TopoART) and            *
*             Hypersphere TopoART (class Hypersphere_TopoART)                                      *
****************************************************************************************************
*                           Created by Marko Tscherepanow, 30 December 2012                        *
***************************************************************************************************/

// Compile and run from the console: dotnet run -p TopoART_sample2.csproj

using System;
using System.IO;
using System.Reflection;
using System.Text.RegularExpressions;
using LibTopoART;
using System.Globalization;

namespace LibTopoART_samples
{
	/// <summary>
	/// Clustering sample using more complex synthetic two-dimensional data. [C#]
	/// <para>
	/// Train TopoART or Hypersphere TopoART with a two-dimensional dataset similar to the one used in "Marko 
	/// Tscherepanow and Sören Riechers (2012). An Incremental On-line Classifier for Imbalanced, Incomplete, and Noisy 
	/// Data. In Proceedings of the European Conference on Artificial Intelligence (ECAI), Workshop on Active and 
	/// Incremental Learning (AIL), pp. 18-23. Montpellier, France." This dataset comprises six clusters (each 
	/// containing 15,000 samples) as well as 10,000 noise samples. These samples were mixed randomly.
	/// </para>
	/// <para>The resulting neural network can be visualised using the R script <c>ShowTopoARTResults.R</c> or the 
	/// R script <c>ShowHypersphereTopoARTResults.R</c>, respectively. Both R scripts are provided in the 
	/// subfolder <c>R</c>.
	/// </para>
	/// </summary>
	class TopoART_sample2
	{
		enum Selector {
			USE_TopoART,
			USE_Fast_TopoART,
			USE_Hypersphere_TopoART
		};

	    static void Main() {
	
			// Dataset
			const string Dataset 		=	"../../../../../data/AIL12-like_dataset.txt";	
			// Destination directory for trained networks
			const string NetworkPath	=	"../../../../../results/networks/";

			const long SampleNumber		=	100000;
			const long InputDimension	=	2;

			// Choose TopoART implementation (see enum Selector above)
			Selector sel = Selector.USE_Fast_TopoART;

			ITopoART ta				=	null;
			string FilePrefix		=	null;
			string FileSuffix		=	null;
			decimal[][] InputArray	=	new decimal[SampleNumber][];
			StreamReader DatasetFile;
			long i;
			string line;
			DateTime start, end;
			TimeSpan time;

			// Set working directory to assembly directory
            Directory.SetCurrentDirectory(Path.GetDirectoryName(new Uri(Assembly.GetEntryAssembly().Location).LocalPath));
			
			// Load dataset
			using(DatasetFile = new StreamReader(File.Open(Dataset, FileMode.Open))) {
				for(i = 0; i < SampleNumber; ++i) {
					line = DatasetFile.ReadLine();
					string[] numbers = Regex.Split(line, @"\s+");
					InputArray[i] = new decimal[InputDimension]; 
					InputArray[i][0] = Decimal.Parse(numbers[1], NumberStyles.Float, System.Globalization.CultureInfo.InvariantCulture);
					InputArray[i][1] = Decimal.Parse(numbers[2], NumberStyles.Float, System.Globalization.CultureInfo.InvariantCulture);
				}
			}

			// Start time measuring
			start = DateTime.Now;
			
			// Initialise a TopoART network with appropriate parameter values
			switch(sel) {
				case Selector.USE_TopoART:
						ta = new TopoART(InputDimension, 2, 0.92m);	
						ta.Beta_sbm	=	0.65m;
						ta.Phi		=	5;
						ta.Tau		=	200;
						FilePrefix 	=	"TopoART";
						FileSuffix	=	"ta";
					break;
				case Selector.USE_Fast_TopoART:
						ta = new Fast_TopoART(InputDimension, 2, 0.92m);	
						ta.Beta_sbm	=	0.65m;
						ta.Phi		=	5;
						ta.Tau		=	200;
						FilePrefix 	=	"Fast_TopoART";
						FileSuffix	=	"fta";
					break;
				case Selector.USE_Hypersphere_TopoART:
						ta = new Hypersphere_TopoART(InputDimension, 2, 0.90m);
						ta.Beta_sbm	=	0.60m;
						ta.Phi		=	5;
						ta.Tau		=	200;
						FilePrefix 	=	"Hypersphere_TopoART";
						FileSuffix	=	"hta";
					break;
			}
					
			// Train network
			for(i = 0; i < SampleNumber; ++i) {
				ta.Learn(InputArray[i]);
				if(i % 1000 == 0) 
					Console.Write(".");
			}
			Console.WriteLine("");
			
			// Determine clusters
			ta.ComputeClusterIDs();

			// Stop time measuring
			end = DateTime.Now;

			// Output the required time
			time = end - start;
			Console.WriteLine("Time for computation: " + time);
			
			// Save network in human readable form
			ta.SaveText(NetworkPath + FilePrefix + "_AIL12-like_dataset.txt");

			// Save network in binary form
			ta.Save(NetworkPath + FilePrefix + "_AIL12-like_dataset." + FileSuffix);
		}
	}
}