'***************************************************************************************************
'             VB sample for the usage of TopoART (classes TopoART and Fast_TopoART) and            *
'             Hypersphere TopoART (class Hypersphere_TopoART)                                      *
'***************************************************************************************************
'                           Created by Marko Tscherepanow, 24 October 2019                         *
'***************************************************************************************************

' Compile and run from the console: dotnet run -p TopoART_sample3.vbproj

Imports System
Imports System.Globalization
Imports System.IO
Imports System.Reflection
Imports System.Text.RegularExpressions
Imports LibTopoART

Namespace LibTopoART_Samples

	''' <summary>
	''' Clustering sample using very noisy synthetic two-dimensional data. [VB]
	''' <para>Train TopoART or Hypersphere TopoART with a two-dimensional dataset similar to the one used in "Marko 
	''' Tscherepanow and Sören Riechers (2012). An Incremental On-line Classifier for Imbalanced, Incomplete, and Noisy 
	''' Data. In Proceedings of the European Conference on Artificial Intelligence (ECAI), Workshop on Active and 
	''' Incremental Learning (AIL), pp. 18-23. Montpellier, France." The dataset applied here comprises 1,000,000 samples 
	''' equally allotted to six clusters (each containing one sixth of the samples). Additionally, 1,000,000 uniformly
	''' distributed random samples are added. Finally, all samples are randomly shuffled.</para>
	''' <para>Due to the randomness involved, the results differ between different runs of this program. However, they  
	''' are qualitatively comparable: The first module creates a coarse clustering of the data while the second module 
	''' refines it to the six clusters of the undisturbed portion of the dataset.</para>
	''' <para>These results show the abilities of TopoART to cope with a high amount of noise data and produce 
	''' stable results independent of the sample order.</para>
	''' <para>The resulting neural network can be visualised using the R script <c>ShowTopoARTResults.R</c> or the 
	''' R script <c>ShowHypersphereTopoARTResults.R</c>, respectively. Both R scripts are provided in the 
	''' subfolder <c>R</c>.</para>
	''' </summary>
	Module TopoART_sample3

		Enum Selector
			USE_TopoART
			USE_Fast_TopoART
			USE_Hypersphere_TopoART
		End Enum
		
		' Choose TopoART implementation (see enum Selector above)
		Const sel = Selector.USE_Fast_TopoART
		
		' Dataset (1,000,000 cluster samples without noise samples)
		Const Dataset 		=	"../../../../../data/AIL12-like_dataset_large_and_noise_free.txt"	
		' Destination directory for trained networks
		Const NetworkPath	=	"../../../../../results/networks/"
		
		' Number of samples in the dataset
		Const DatasetSampleNumber	=	1000000
		
		' Number of noise samples to be added
		Const NoiseSampleNumber		=	1000000
		
		Const SampleNumber			=	DatasetSampleNumber + NoiseSampleNumber
		Const InputDimension		=	2
		
		' Common network parameters
		Const ModuleNumber	=	2
		Const rho_a			=	0.92D
		Const beta_sbm		=	0.65D
		Const tau			=	2000
		
		Sub Main(args As String())
			Dim ta As ITopoART = Nothing
			Dim FilePrefix As String = Nothing 
			Dim FileSuffix As String = Nothing
			Dim InputArray(SampleNumber)() As Decimal
			Dim line As String
			Dim StartTime, StopTime As DateTime
			Dim time As TimeSpan
			Dim rnd As New Random
			
			' Set working directory to assembly directory
			Directory.SetCurrentDirectory(Path.GetDirectoryName(New Uri(Assembly.GetEntryAssembly().Location).LocalPath))
			
			Console.WriteLine("Prepare training data")
			
			' Load dataset
			Using DatasetFile As New StreamReader(File.Open(Dataset, FileMode.Open)) 
				For i As Integer = 0 To DatasetSampleNumber - 1
					line = DatasetFile.ReadLine()
					Dim Numbers() As String = Regex.Split(line, "\s+")
					Dim SingleInput(InputDimension) As Decimal 
					SingleInput(0) = Decimal.Parse(numbers(1), NumberStyles.Float, CultureInfo.InvariantCulture)
					SingleInput(1) = Decimal.Parse(numbers(2), NumberStyles.Float, CultureInfo.InvariantCulture)
					InputArray(i) = SingleInput
				Next
			End Using
			
			' Add noise samples
			For i As Integer = 0 To NoiseSampleNumber - 1
				Dim SingleInput(InputDimension) As Decimal 
				SingleInput(0) = rnd.NextDouble()
				SingleInput(1) = rnd.NextDouble()
				InputArray(i + DatasetSampleNumber) = SingleInput
			Next
			
			' Shuffle data
			For i As Integer = 0 To SampleNumber - 1
				Dim NewIndex As Integer = Math.Round(rnd.NextDouble() * (SampleNumber - 1))
				Dim TmpInput() As Decimal = InputArray(i)
				InputArray(i) = InputArray(NewIndex)
				InputArray(NewIndex) = TmpInput
			Next
			
			' Start time measuring
			StartTime = Now
			
			' Initialise a TopoART network with appropriate parameter values
			Select sel 
				Case Selector.USE_TopoART
						ta = New TopoART(InputDimension, ModuleNumber, rho_a)	
					
						FilePrefix 	=	"TopoART"
						FileSuffix	=	"ta"
				Case Selector.USE_Fast_TopoART
						ta = New Fast_TopoART(InputDimension, ModuleNumber, rho_a)
						FilePrefix 	=	"Fast_TopoART"
						FileSuffix	=	"fta"
				Case Selector.USE_Hypersphere_TopoART
						ta = New Hypersphere_TopoART(InputDimension, ModuleNumber, rho_a)
						FilePrefix 	=	"Hypersphere_TopoART"
						FileSuffix	=	"hta"
			End Select
			
			' Set remaining common parameters
			ta.Beta_sbm	=	beta_sbm
			ta.Tau		=	tau
			ta.Phis		=	New Long() {28, 15}
						
			' Train network
			For i As Long = 0 To SampleNumber - 1
				ta.Learn(InputArray(i))
				If (i Mod 20000) = 0 
					Console.Write(".")
				End If
			Next
			Console.WriteLine("")
			
			' Determine clusters
			ta.ComputeClusterIDs()
			
			' Stop time measuring
			StopTime = Now

			' Output the required time
			time = StopTime - StartTime
			Console.WriteLine("Time for computation: " & time.ToString())
			
			' Save network in human readable form
			ta.SaveText(NetworkPath + FilePrefix + "_AIL12-like_dataset_large_and_noisy.txt")

			' Save network in binary form
			ta.Save(NetworkPath + FilePrefix + "_AIL12-like_dataset_large_and_noisy." + FileSuffix)
		End Sub

	End Module

End Namespace
